from edsteva.viz.dashboards.estimates_dashboard import estimates_dashboard
from edsteva.viz.dashboards.predictor_dashboard.wrapper import predictor_dashboard

__all__ = [
    "predictor_dashboard",
    "estimates_dashboard",
]
